#
# This script can be used to produce prototype HMMSets for
# PLAINHS, TIEDHS and DISCRETEHS systems
#

# Global Variables
$nStates=0;                         # Number of active HMM states
$nStreams=0;                        # Number of streams
$vecSize=0;                         # Size of feature vector
$pKind="";                          # Parameter kind
$cKind="";                          # Covariance kind
$hsKind="";                         # System kind
$outDir="";                         # Output HMMSet directory
$configParams;                      # Global store of config parameters


#***************************** START MAIN ***********************************
$|=1; #Force buffer flush on STDOUT


if ($ARGV[0]){
    &ReadPCF();
    &SetVars();
    &WriteHSet();
}else{
    print "USAGE: MakeProtoHMMSet ConfigFile\n";
}

#******************************* END MAIN *********************************

#************************ Util Functions **********************************

#-------------------------------------
# ReadPCF: Reads the Proto Config File 
#-------------------------------------
sub ReadPCF {

local($validData,$param,$val)=0;

while(<>){
    if(/\<ENDsys_setup\>/){
	$validData=0;
    }
    if($validData){
	($param,$val)=split(/ *: */, $_);
	if (($param =~ /hmmList/)||($param =~ /outDir/)||($param =~ /parmKind/)) {
	}elsif ($param =~ /mixes/){
	    @mixes=split(/ +/, $_);
	}elsif ($param =~ /sWidths/){
	    @sWidths=split(/ +/, $_);
	}else{
	    $val =~ tr/A-Z/a-z/;
	}
        chop($val);
        $configParams{$param}=$val;
        write;
    }
    if(/\<BEGINsys_setup\>/ || /\<BEGINtool_steps\>/){
	$validData=1;
    }
}
format STDOUT_TOP =
 Proto Config File Read
 ======================
Parameter         Value
-----------------------
.
format STDOUT=
@<<<<<<<<<<<<<<<<<<<@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$param,$val
.
}

#-----------------------------------------
# SetVars: Set variable from config params
#-----------------------------------------
sub SetVars {

$nStates = $configParams{"nStates"};
($nStates >= 1) || die "nStates must be >= 1";
$nStreams = $configParams{"nStreams"};
($nStreams =~ /^[1234]/) || die "nStreams must be 1,2,3 or 4";
$vecSize = $configParams{"vecSize"};
($vecSize >= 1) || die "vecSize must be >= 1";
$hsKind = $configParams{"hsKind"};
($hsKind =~ /^[pPtTdD]/) || die "hsKind must be P, T or D";
$cKind = $configParams{"covKind"};
($cKind =~ /^[dDfF]/) || die "covKind must be F or D";
$pKind = $configParams{"parmKind"};
if ($hsKind=~/^[dD]/){
 ($pKind=~/[vV]$/) || die "If hsKind is D then parmKind must have _V appended";
}
if ($hsKind=~/^[pPtT]/){
    if($pKind=~/[vV]$/){
	die "If hsKind is not D then parmKind must not have _V appended";
    }
}
$outDir = $configParams{"outDir"};
opendir(OUTDIR,$outDir) || die "Can't open $outDir";
$hmmList = $configParams{"hmmList"};
(-f $hmmList) || die "Cannot find HMM list file $hmmList";
}


#************************ HMM Functions **********************************

#------------------------------------
# WriteGlobOpts: Write global options
#------------------------------------
sub WriteGlobOpts {
    local($fleHandle,$hmmName)=@_;

    printf($fleHandle "  ~o <VECSIZE> %d <$pKind> <STREAMINFO> %d ",$vecSize,$nStreams);
    for ($i=1; $i<=$nStreams; $i++) {
	printf($fleHandle "%d ",$sWidths[$i]);
    }
    printf($fleHandle "\n");
    (($hsKind=~/^[tT]/)&&($fleHandle=~/MACRO/)) || printf($fleHandle "  ~h \"$hmmName\"\n");
}

#-------------------------------------------------------
# WriteDiagCMixtures: write a diagonal covariance mixture
#-------------------------------------------------------
sub WriteDiagCMixtures {
    local($streamNum)=@_;
    local($i,$j,$k,$mixWght)=0;

    $mixWght=1/$mixes[$streamNum];

    for ($i=1; $i<=$mixes[$streamNum]; $i++){
	printf(PROTO "  <MIXTURE> %d %1.4f\n",$i,$mixWght);
	printf(PROTO "    <MEAN> %d\n",$sWidths[$streamNum]);
	printf(PROTO "      ");
	for ($j=1; $j<=$sWidths[$streamNum]; $j++){
	    printf(PROTO "0.0 ");
	}
	printf(PROTO "\n");
	printf(PROTO "    <VARIANCE> %d\n",$sWidths[$streamNum]);
	printf(PROTO "      ");
	for ($k=1; $k<=$sWidths[$streamNum]; $k++){
	    printf(PROTO "1.0 ");
	}
	printf(PROTO "\n");
    }
}

#-------------------------------------------------------
# WriteFullCMixtures: write the full covariance mixtures
#-------------------------------------------------------
sub WriteFullCMixtures {
    local($streamNum)=@_;
    local($i,$j,$k,$mixWght,$tmpVecSize)=0;

    $mixWght=1/$mixes[$streamNum];
    $tmpVecSize=$sWidths[$streamNum];

    for ($i=1; $i<=$mixes[$streamNum]; $i++){
	printf(PROTO "  <MIXTURE> %d %1.4f\n",$i,$mixWght);
	printf(PROTO "    <MEAN> %d\n",$sWidths[$streamNum]);
	printf(PROTO "      ");
	for ($j=1; $j<=$sWidths[$streamNum]; $j++){
	    printf(PROTO "0.0 ");
	}
	printf(PROTO "\n");
	printf(PROTO "    <INVCOVAR> %d\n",$sWidths[$streamNum]);
	while ($tmpVecSize>=1){
	    for ($k=1; $k<=$tmpVecSize; $k++){
		if ($k==1){
		    printf(PROTO "1.0 ");
		}else{
		    printf(PROTO "0.0 ");
		}
	    }
	    printf(PROTO "\n");
	    $tmpVecSize--;
	}
	$tmpVecSize=$sWidths[$streamNum];
    }
}


#----------------------------------------------
# WriteDProbs: Write the discrete probabilities
#----------------------------------------------
sub WriteDProbs {
    local($s)=@_;

    printf(PROTO "      <DPROB> %d*$mixes[$s]\n",-2371.8*log(1/$mixes[$s]));
}


#----------------------------------------------
# WriteStates: Write the contents of the states
#----------------------------------------------
sub WriteStates {
    local($i,$j)=0;

    for ($i=1; $i<=$nStates; $i++){
	printf(PROTO "  <STATE> %d \n",$i+1);
	printf(PROTO "  <NUMMIXES> ");
        for ($k=1; $k<=$nStreams; $k++){
	    printf(PROTO "%d ",$mixes[$k]);
	}
	printf(PROTO "\n");
	for ($j=1; $j<=$nStreams; $j++){
	    printf(PROTO "  <STREAM> %d\n",$j);
	    if ($hsKind =~ /^[Dd]/){
		&WriteDProbs($j);
	    }elsif ($hsKind =~ /^[pP]/){
		if ($cKind =~ /^[dD]/){
		    &WriteDiagCMixtures($j);
		}else{
		    &WriteFullCMixtures($j);
		}
	    }
	}
    }
}

#-----------------------------------------------------------
# WriteTransMat: Write the contents of the transition matrix
#-----------------------------------------------------------
sub WriteTransMat {
    local($i,$j)=0;

    printf(PROTO "  <TRANSP> %d\n",$nStates+2);
    for ($i=1; $i<=$nStates+2; $i++){
	for ($j=1; $j<=$nStates+2; $j++){
	    if (($i==1)&&($j==2)){
		printf (PROTO "   1.000e+0");
	    }elsif (($i==$j)&&($i!=1)&&($i!=$nStates+2)){
		printf (PROTO "   6.000e-1");
	    }elsif ($i==($j-1)){
		printf (PROTO "   4.000e-1");
	    }else{
		printf (PROTO "   0.000e+0");
	    }
	}
	printf (PROTO "\n");
    }	    
}

#----------------------------------------
# WriteHMM: Write the contents of the HMM
#----------------------------------------
sub WriteHMM {
    local($hmmName)=@_;
    local($i)=0;

    open(PROTO, ">$outDir/$hmmName") || die "Cannot open $outDir/$hmmName for writing";
    &WriteGlobOpts(PROTO,$hmmName);
    printf(PROTO "<BEGINHMM>\n");
    printf(PROTO "  <NUMSTATES> %d\n",$nStates+2);
    &WriteStates();
    &WriteTransMat();
    printf(PROTO "<ENDHMM>\n");
    close(PROTO);
}

#------------------------------------------
# WriteHSet: Write the contents of the HSet
#------------------------------------------
sub WriteHSet {

    open(HMMLIST, $hmmList) || die "Cannot open $hmmList for reading";

    printf "\nWriting HMMSet\n\n";
    while (<HMMLIST>) {
	chop($_);
	printf "Writing HMMDef to $outDir/$_\n";
	&WriteHMM($_);
    }
    close(HMMLIST);
}


